#define USECHRONO
// #define HAVE_MPI
#include "./eval.hpp"
#include <cassert>
#include <climits>
#include <cstdlib>

// using namespace aed;
//---:---<*>---:---<*>---:---<*>---:---<*>---:---<*>
// test if XPROBE is inside simplex XVRTX
bool inside(MatrixXd &xvrtx, VectorXd &xprobe) {
  int ndim = xprobe.size();
  VectorXd rhs(ndim+1),alpha(ndim+1);
  MatrixXd C(ndim+1,ndim+1);
  C.block(0,0,ndim,ndim+1) = xvrtx;
  for (int k=0; k<ndim+1; k++) C(ndim,k)=1.0;
  rhs.segment(0,ndim) = xprobe;
  rhs(ndim) = 1.0;
  alpha = C.lu().solve(rhs);
  for (int j=0; j<ndim+1; j++)
    if (alpha(j)<0.0 || alpha(j)>1) return false;
  return true;
}

// Returns true if simplex B is included in simplex A
bool simplexinc(MatrixXd &XA, MatrixXd &XB) {
  //C¿ Check if each node of the probe tetra (XB) is inside
  //tnhe XA tetra
  int ndim = XA.rows();
  assert(XA.cols()==ndim+1);
  assert(XB.rows()==ndim);
  assert(XB.cols()==ndim+1);
  VectorXd xbj(ndim);
  for (int j=0; j<ndim+1; j++) {
    xbj = XB.col(j);
    if (!inside(XA,xbj)) return false;
  }
  return true;
}

// Version that avoids race condition using a CRITICAL section
void getcover_critical(vector<MatrixXd> &tetras,
                       vector<MatrixXd> &cover,
                       vector<MatrixXd> &notcover) {
  cover.clear();
  int ntet = tetras.size();
#pragma omp parallel for  
  for (int j=0; j<ntet; j++) {
    int incover=1;
    for (int k=0; k<ntet; k++) {
      // If TJ is included in some TK then it doesn't belong to COVER
      if (j!=k && simplexinc(tetras[k],tetras[j])) {
        incover=0; break;
      }
    }
    // The race condition may happen when accessing the
    // shared COVER and NOTCOVER structures
#pragma omp critical
    {
      if (incover) cover.push_back(tetras[j]);
      else notcover.push_back(tetras[j]);
    }
  }
}

// Avoid the race condition by using a separate storage area
// for each thread. 
void getcover_tidstor(vector<MatrixXd> &tetras,
                      vector<MatrixXd> &cover,
                      vector<MatrixXd> &notcover) {
  cover.clear();
  // We have to call this in order to get the number of
  // threads that will be forked in the parallel region
  int nthreads = omp_get_max_threads();
  // Dimension the private storage areas
  vector< vector<MatrixXd> >
    cover_tid(nthreads),
    notcover_tid(nthreads);
  int ntet = tetras.size();
  // Fork the parallel region
#pragma omp parallel
  {
    // get my thread ID in order to acces my own separate storage area 
    int tid = omp_get_thread_num();
    // Loop in parallel
#pragma omp for    
    for (int j=0; j<ntet; j++) {
      // Check if tetra Tj is inside some other Tk, with k!=j
      int incover=1;
      for (int k=0; k<ntet; k++) {
        // If TJ is included in some TK then it doesn't belong to COVER
        if (j!=k && simplexinc(tetras[k],tetras[j])) {
          incover=0; break;
        }
      }
      // Push to the appropriate private storage area
      if (incover) cover_tid[tid].push_back(tetras[j]);
      else notcover_tid[tid].push_back(tetras[j]);
    }
  }
  // Now OUTSIDE the parallel region gather all the private
  // areas into a global storage area
  for (int tid=0; tid<nthreads; tid++) {
    for (auto q:cover_tid[tid]) cover.push_back(q);
    for (auto q:notcover_tid[tid]) notcover.push_back(q);
  }
}

// Use locks for each of the shared structures. COVER_LOCK
// for the COVER structure and NOTCOVER_LOCK for the
// NOTCOVER structure.
void getcover_lock(vector<MatrixXd> &tetras,
                   vector<MatrixXd> &cover,
                   vector<MatrixXd> &notcover) {
  cover.clear();
  int ntet = tetras.size();
  // Declare and initialize the locks
  omp_lock_t cover_lock,notcover_lock;
  omp_init_lock(&cover_lock);
  omp_init_lock(&notcover_lock);
#pragma omp parallel for
  for (int j=0; j<ntet; j++) {
    int incover=1;
    for (int k=0; k<ntet; k++) {
      // If TJ is included in some TK then it doesn't belong to COVER
      if (j!=k && simplexinc(tetras[k],tetras[j])) {
        incover=0; break;
      }
    }
    if (incover) {
      // Push to the COVER structure with a semaphore
      omp_set_lock(&cover_lock);
      cover.push_back(tetras[j]);
      omp_unset_lock(&cover_lock);
    } else {
      // Push to the NOTCOVER structure with a semaphore
      omp_set_lock(&notcover_lock);
      notcover.push_back(tetras[j]);
      omp_unset_lock(&notcover_lock);
    }
  }
  // Destroy the locks
  omp_destroy_lock(&cover_lock);
  omp_destroy_lock(&notcover_lock);
}

int main() {
  int seed0=123456;
  for (int j=0; j<10; j++) {
    // check_getcover() checks a solution w.r.t. a reference
    // solution
    int seed = seed0+j;
    printf("checking getcover_critical %d\n",
           check_getcover(seed,getcover_critical));
    printf("checking getcover_tidstor %d\n",
           check_getcover(seed,getcover_tidstor));
    printf("checking getcover_lock %d\n",
           check_getcover(seed,getcover_lock));
  }
  return 0;
}
